import pandas as pd
from dataclasses import dataclass
from abc import ABC, abstractmethod

from df_and_order.helpers import build_class_instance


TRANSFORM_STEP_MODULE_PATH_KEY = 'module_path'
TRANSFORM_STEP_PARAMS_KEY = 'params'
TRANSFORM_STEP_NEED_CACHE_KEY = 'needs_cache'


@dataclass
class DfTransformStepConfig:
    """
    Dataclass for storing module path of some DfTransformStep
    as well as its init parameters.
    """
    module_path: str
    params: dict

    @staticmethod
    def from_dict(transform_dict: dict):
        module_path = transform_dict[TRANSFORM_STEP_MODULE_PATH_KEY]
        params = transform_dict.get(TRANSFORM_STEP_PARAMS_KEY) or {}
        config = DfTransformStepConfig(module_path=module_path,
                                       params=params)
        return config

    def to_dict(self) -> dict:
        step_dict = {
            TRANSFORM_STEP_MODULE_PATH_KEY: self.module_path,
        }

        if len(self.params):
            step_dict[TRANSFORM_STEP_PARAMS_KEY] = self.params

        return step_dict


class DfTransformStep(ABC):
    """
    Encapsulates logic of some dataframe transformation.
    Every subclass must implement 'transform' method with
    custom logic.
    """
    @staticmethod
    def build_transform(config: DfTransformStepConfig):
        """
        Creates DfTransformStep instance out of config.

        Parameters
        ----------
        config: DfTransformStepConfig
            Contains information about DfTransformStep subclass
            and parameters to init it with.

        Returns
        -------
        DfTransformStep instance.
        """
        params = config.params

        transform = build_class_instance(module_path=config.module_path,
                                         init_params=params)
        return transform

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass



from typing import Optional
import os
import yaml

from df_and_order.df_transform import DfTransformConfig

DF_ID_KEY = 'df_id'
DF_INITIAL_FORMAT_KEY = 'initial_df_format'
DF_TRANSFORMED_FORMAT_KEY = 'transformed_df_format'
METADATA_KEY = 'metadata'
TRANSFORMS_KEY = 'transforms'


class DfConfig:
    """
    Interface for working with a serialized dataframe config file.

    Init parameters
    ----------
    df_id: str
        Unique identifier of your dataframe.
    dir_path: str
        Location of a folder where config file should be read from.
    """
    def __init__(self, df_id: str, dir_path: str):
        self._dir_path = dir_path
        with open(self._config_path()) as config_file:
            config_dict = yaml.safe_load(config_file)

        df_check = config_dict[DF_ID_KEY] == df_id
        assert df_check, "config_dict doesn't belong to requested dataframe"

        self._config_dict = config_dict

    @staticmethod
    def config_exists(dir_path: str) -> bool:
        """
        Checks whether a dataframe config file exists at the provided path.

        Parameters
        ----------
        dir_path: str
            Location of a folder where config file should be read from.

        Returns
        -------
        True if config file exists, False otherwise
        """
        path_to_check = DfConfig._config_path_for(dir_path=dir_path)
        return os.path.exists(path_to_check)

    def _config_path(self) -> str:
        """
        Gets a path to the config using the given dir_path
        """
        return self._config_path_for(dir_path=self._dir_path)

    @staticmethod
    def _config_path_for(dir_path: str) -> str:
        """
        Gets a path to the config using the given dir_path
        """
        return DfConfig._path_for_file(dir_path=dir_path, filename='df_config.yaml')

    @staticmethod
    def _path_for_file(dir_path: str, filename: str) -> str:
        return os.path.join(dir_path, filename)

    @staticmethod
    def create_config(dir_path: str,
                      df_id: str,
                      initial_df_format: str,
                      transformed_df_format: str,
                      metadata: Optional[dict] = None,
                      transform: Optional[DfTransformConfig] = None):
        """
        Creates config file at the given location.

        Parameters
        ----------
        dir_path: str
            Location of a folder where config file should be saved to
        df_id: str
            Unique identifier of your dataframe.
        initial_df_format: str
            Format extension for an initial dataframe before any transformation.
        transformed_df_format: str
            # TODO: let a transformation decide in which format to save a dataframe
            Format extension for a transformed dataframe.
        metadata: optional dict
            Any information you want to store in the config file as a reference or for future use.
        transform: DfTransformConfig
            Object that describes all the transformation steps required.

        Returns
        -------
        Nothing, but creates the config in the filesystem.
        """
        assert not DfConfig.config_exists(dir_path=dir_path), f"Config for df {df_id} already exists"

        config_path = DfConfig._config_path_for(dir_path=dir_path)
        config_dict = {
            DF_ID_KEY: df_id,
            DF_INITIAL_FORMAT_KEY: initial_df_format,
            DF_TRANSFORMED_FORMAT_KEY: transformed_df_format,
        }

        if metadata:
            config_dict[METADATA_KEY] = metadata

        if transform:
            transform_id, transform_dict = transform.to_dict()
            config_dict[TRANSFORMS_KEY] = {
                transform_id: transform_dict
            }

        DfConfig._save_at_path(config_dict=config_dict,
                               config_path=config_path)

    @property
    def initial_df_format(self) -> str:
        return self._config_dict[DF_INITIAL_FORMAT_KEY]

    @property
    def transformed_df_format(self) -> str:
        return self._config_dict[DF_TRANSFORMED_FORMAT_KEY]

    def transforms_by(self, transform_id: str) -> DfTransformConfig:
        """
        Gets the transform representation from the config file.

        Parameters
        ----------
        transform_id: str
            Unique identifier of a transform

        Returns
        -------
        DfTransformConfig that describes all the transformation steps.
        """
        maybe_transforms_dict = self._config_dict.get(TRANSFORMS_KEY)
        assert maybe_transforms_dict, "Config contains no transforms"
        maybe_transform_dict = maybe_transforms_dict.get(transform_id)
        assert maybe_transform_dict, f"Requested transform {transform_id} was not found"
        result = DfTransformConfig.from_dict(transform_id=transform_id,
                                             transform_dict=maybe_transform_dict)

        return result

    def register_transform(self,
                           transform: DfTransformConfig,
                           filename: str):
        """
        Adds a new transform to the config file if possible.

        Parameters
        ----------
        transform: DfTransformConfig
            Object that describes all the transformation steps required.
        filename: str
            What a filename for the transformation would be to look for already existing one.

        Returns
        -------
        Nothing, but updates the config in the filesystem.
        """
        transform_id, transform_dict = transform.to_dict()
        maybe_transforms = self._config_dict.get(TRANSFORMS_KEY, {})
        if maybe_transforms and maybe_transforms.get(transform_id):
            full_filename = f'{filename}.{self._config_dict[DF_TRANSFORMED_FORMAT_KEY]}'
            already_cached = os.path.exists(self._path_for_file(dir_path=self._dir_path, filename=full_filename))
            assert not already_cached, f"Result of the transform {transform_id} already exists."

        maybe_transforms[transform_id] = transform_dict
        self._config_dict[TRANSFORMS_KEY] = maybe_transforms
        self._save()

    def _save(self):
        """
        Serializes config to the disk
        """
        self._save_at_path(config_dict=self._config_dict,
                           config_path=self._config_path())

    @staticmethod
    def _save_at_path(config_dict: dict, config_path: str):
        with open(config_path, 'w') as config_file:
            yaml.dump(config_dict, config_file)



import importlib
from typing import Optional


def build_class_instance(module_path: str, init_params: Optional[dict] = None):
    """
    Create an object instance from absolute module_path string.

    Parameters
    ----------
    module_path: str
        Full module_path that is valid for your project or some external package.
    init_params: optional dict
        These parameters will be used as init parameters for the given type.

    Returns
    -------
    Some object instance
    """
    class_ = get_type_from_module_path(module_path=module_path)
    result = class_(**(init_params or {}))
    return result


def get_type_from_module_path(module_path: str):
    """
    Takes full module path and takes the class
    name to convert it to Python type.

    Parameters
    ----------
    module_path: str
        Full module_path that is valid for your project or some external package.

    Returns
    -------
    Python type.
    """
    components = module_path.split('.')
    class_name = components[-1]
    module_name = '.'.join(components[:-1])

    module = importlib.import_module(module_name)
    class_type = getattr(module, class_name)
    return class_type



import os
import pandas as pd
from typing import Optional, Dict, List

from df_and_order.df_transform import DfTransformConfig
from df_and_order.df_config import DfConfig
from df_and_order.df_cache import DfCache
from df_and_order.df_transform_step import DfTransformStepConfig, DfTransformStep


class DfReader:
    """
    Reads a dataframe from the disk using a config file.
    Helps you to organize and document your dataframes
    for easy access and reproducibility.

    Parameters
    ----------
    dir_path: str
        Absolute path where your dataframe is supposed to be saved.
    format_to_cache_map: dict
        To be able to work with various formats DfReader must know
        how to read&save them. Provide a map where a format extension is used as a key
        and DfCache instance as a value. See DfCache class documentation for the details.
    """
    def __init__(self,
                 dir_path: str,
                 format_to_cache_map: Dict[str, DfCache]):
        self._dir_path = dir_path
        self._format_to_cache_map = format_to_cache_map

    def df_exists(self,
                  df_id: str,
                  transform_id: Optional[str] = None) -> bool:
        """
        Checks whether a dataframe file exists at the provided path.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            If you want to check whether a transformed version of your dataframe
            is persisted on the disk you may pass its unique identifier that
            matches the one in the config file.

        Returns
        -------
        True if the dataframe exists, False otherwise
        """
        df_dir_path = self._df_dir_path(df_id=df_id)
        if not DfConfig.config_exists(dir_path=df_dir_path):
            return False

        df_config = self._get_config(df_id=df_id)
        df_path = self._df_path(df_config=df_config,
                                df_id=df_id,
                                transform_id=transform_id)
        return os.path.exists(path=df_path)

    def create_df_config(self,
                         df_id: str,
                         initial_df_format: str,
                         transformed_df_format: str,
                         metadata: Optional[dict] = None,
                         transform: Optional[DfTransformConfig] = None):
        """
        Just a forwarding to DfConfig method, see docs in DfConfig.
        """
        DfConfig.create_config(dir_path=self._df_dir_path(df_id=df_id),
                               df_id=df_id,
                               initial_df_format=initial_df_format,
                               transformed_df_format=transformed_df_format,
                               metadata=metadata,
                               transform=transform)

    def register_transform(self,
                           df_id: str,
                           transform: DfTransformConfig):
        """
        Forms a filename for the given dataframe and adds a new transform to the config file if possible.
        In general it's just a forwarding to DfConfig method, see docs in DfConfig.
        """
        df_config = self._get_config(df_id=df_id)
        filename = self.df_filename(df_config=df_config,
                                    df_id=df_id,
                                    transform_id=transform.transform_id)
        df_config.register_transform(transform=transform,
                                     filename=filename)

    def read(self,
             df_id: str,
             transform_id: Optional[str] = None,
             transform: Optional[DfTransformConfig] = None) -> pd.DataFrame:
        """
        Reads a dataframe from the disk. If you want a transformed version of your dataframe,
        but it's still not persisted, it first creates it and then reads it into memory.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            Unique identifier of the desired transform.
        transform: optional DfTransformConfig
            Object that describes all the transformation steps required.

        Returns
        -------

        """
        if transform_id and transform:
            raise AttributeError('Provide either transform_id or transform_config')

        if transform:
            transform_id = transform.transform_id

        df_config = self._get_config(df_id=df_id)

        if transform_id:
            if not transform:
                transform = df_config.transforms_by(transform_id=transform_id)

            return self._read_transformed(df_id=df_id,
                                          transform=transform,
                                          df_config=df_config)
        else:
            return self._read_initial(df_id=df_id,
                                      df_config=df_config)

    def _read_initial(self,
                      df_id: str,
                      df_config: DfConfig) -> pd.DataFrame:
        """
        Reads the original dataframe from the disk
        """
        return self._read_df(df_id=df_id,
                             df_format=df_config.initial_df_format,
                             df_config=df_config)

    def _read_transformed(self,
                          df_id: str,
                          transform: DfTransformConfig,
                          df_config: DfConfig) -> pd.DataFrame:
        """
        Reads the transformed dataframe from the disk or creates it if needed.
        """
        transform_id = transform.transform_id
        transformed_df_exists = self.df_exists(df_id=df_id,
                                               transform_id=transform_id)
        df_format = df_config.transformed_df_format
        if transformed_df_exists:
            df = self._read_df(df_config=df_config,
                               df_id=df_id,
                               df_format=df_format,
                               transform_id=transform_id)

            if transform.in_memory_steps:
                df = DfReader._apply_transform_steps(df=df,
                                                     steps=transform.in_memory_steps)

            return df

        self.register_transform(df_id=df_id,
                                transform=transform)

        df = self._read_initial(df_id=df_id,
                                df_config=df_config)

        if transform.in_memory_steps:
            df = DfReader._apply_transform_steps(df=df,
                                                 steps=transform.in_memory_steps)

        if transform.permanent_steps:
            df = DfReader._apply_transform_steps(df=df,
                                                 steps=transform.permanent_steps)

            df_path = self._df_path(df_config=df_config,
                                    df_id=df_id,
                                    transform_id=transform_id)
            df_cache = self._get_df_cache(df_format=df_format)
            df_cache.save(df=df, path=df_path)

        return df

    @staticmethod
    def _apply_transform_steps(df: pd.DataFrame,
                               steps: List[DfTransformStepConfig]) -> pd.DataFrame:
        """
        Applies all the steps for a transformation on the given dataframe.

        Parameters
        ----------
        df: pd.DataFrame
            Initial dataframe to perform transformations on.
        steps: list of DfTransformStepConfig
            List of objects that represent a step of the whole transformation.

        Returns
        -------
        pd.DataFrame, fully transformed initial dataframe
        """
        for step in steps:
            transform = DfTransformStep.build_transform(config=step)
            df = transform.transform(df=df)

        return df

    def _read_df(self,
                 df_id: str,
                 df_format: str,
                 df_config: DfConfig,
                 transform_id: Optional[str] = None) -> pd.DataFrame:
        """
        General method for reading a dataframe from the disk.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        df_format: str
            Format in which dataframe was saved.
        df_config: DfConfig
            Interface for working with a serialized dataframe config file.
        transform_id: optional str
            Unique identifier of the desired transform.

        Returns
        -------
        pd.DataFrame, the requested dataframe
        """
        df_path = self._df_path(df_config=df_config,
                                df_id=df_id,
                                transform_id=transform_id)

        df_cache = self._get_df_cache(df_format=df_format)
        df = df_cache.load(path=df_path)

        return df

    def _df_path(self,
                 df_config: DfConfig,
                 df_id: str,
                 transform_id: Optional[str] = None) -> str:
        """
        Forms a path to the dataframe.

        Parameters
        ----------
        df_config: DfConfig
            Interface for working with a serialized dataframe config file.
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            Unique identifier of the desired transform.

        Returns
        -------
        str, absolute path to the dataframe
        """
        filename = DfReader.df_filename(df_config=df_config,
                                        df_id=df_id,
                                        transform_id=transform_id)
        result = self._df_dir_path(df_id=df_id, filename=filename)

        return result

    @staticmethod
    def df_filename(df_config: DfConfig,
                    df_id: str,
                    transform_id: Optional[str] = None):
        """
        Forms a filename for the dataframe

        Parameters
        ----------
        df_config: DfConfig
            Interface for working with a serialized dataframe config file.
        df_id: str
            Unique identifier of your dataframe.
        transform_id: optional str
            Unique identifier of the desired transform.

        Returns
        -------
        str, a filename for the dataframe.
        """
        if transform_id:
            return f'{transform_id}_{df_id}.{df_config.transformed_df_format}'

        return f'{df_id}.{df_config.initial_df_format}'

    def _df_dir_path(self, df_id: str, filename: Optional[str] = None) -> str:
        """
        Forms an absolute file for a directory where you can find your dataframe or any other file.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.
        filename: optional str
            If you want to get a path for some particular file, provide its filename.

        Returns
        -------
        str, absolute path to the desired item
        """
        path = self._df_dir_path_for(dir_path=self._dir_path,
                                     df_id=df_id)

        if filename:
            path = os.path.join(path, filename)

        return path

    @staticmethod
    def _df_dir_path_for(dir_path: str,
                         df_id: str) -> str:
        return os.path.join(dir_path, df_id)

    def _get_config(self, df_id: str) -> DfConfig:
        """
        Gets config object for the given dataframe.

        Parameters
        ----------
        df_id: str
            Unique identifier of your dataframe.

        Returns
        -------
        DfConfig instance.
        """
        result = self._get_config_for(dir_path=self._df_dir_path(df_id=df_id),
                                      df_id=df_id)
        return result

    @staticmethod
    def _get_config_for(dir_path: str, df_id: str) -> DfConfig:
        """
        Gets config object for the given dataframe at the given path.

        Parameters
        ----------
        dir_path: str
            Absolute path to the dir when the dataframe is located.
        df_id: str
            Unique identifier of your dataframe.

        Returns
        -------

        """
        result = DfConfig(df_id=df_id,
                          dir_path=dir_path)
        return result

    def _get_df_cache(self, df_format: str) -> DfCache:
        """
        Gives proper DfCache subclass for the given data format

        Parameters
        ----------
        df_format: str
            Format extension you want to use to save/load a dataframe.

        Returns
        -------
        DfCache instance.
        """
        df_cache = self._format_to_cache_map.get(df_format)
        if not df_cache:
            raise ValueError(f'Unknown df_format, df_cache was not provided: {df_format}')
        return df_cache



import pandas as pd
from typing import List

from df_and_order.df_transform_step import DfTransformStep


class DropColsTransformStep(DfTransformStep):
    """
    Example built-in transform that simply drops some
    undesired columns from a dataframe.
    """
    def __init__(self, cols_to_drop: List[str]):
        super().__init__()

        self._cols_to_drop = cols_to_drop

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.drop(self._cols_to_drop, axis=1)



from typing import List, Optional, Tuple

from df_and_order.df_transform_step import DfTransformStepConfig


TRANSFORM_ID_KEY = 'transform_id'
TRANSFORM_IN_MEMORY_KEY = 'in_memory'
TRANSFORM_PERMANENT_KEY = 'permanent'


class DfTransformConfig:
    """
    Describes how a transformation should be performed.
    For any transformation one or many steps can be used.
    Those steps can be performed in memory only or their result
    can be serialized to the disk. Either in_memory or permanent or both steps
    should be provided in the init method.

    Parameters
    ----------
    transform_id: optional str
        Unique identifier of a transform.
    in_memory_steps: list of DfTransformStepConfig
        Those steps that are performed every time we read a dataframe from the disk.
    permanent_steps: list of DfTransformStepConfig
        Those steps result of which is persisted on the disk for future access.
    """
    def __init__(self,
                 transform_id: str,
                 in_memory_steps: Optional[List[DfTransformStepConfig]] = None,
                 permanent_steps: Optional[List[DfTransformStepConfig]] = None):
        assert in_memory_steps or permanent_steps, "Provide at least one type of transformations"

        self._transform_id = transform_id
        self._in_memory_steps = in_memory_steps
        self._permanent_steps = permanent_steps

    @property
    def transform_id(self) -> str:
        return self._transform_id

    @property
    def in_memory_steps(self) -> Optional[List[DfTransformStepConfig]]:
        return self._in_memory_steps

    @property
    def permanent_steps(self) -> Optional[List[DfTransformStepConfig]]:
        return self._permanent_steps

    def to_dict(self) -> Tuple[str, dict]:
        """
        Converts an instance to serializable dictionary

        Returns
        -------
        tuple of str and dict, transform_id and transform's dictionary representation.
        """
        result = {}

        if self._in_memory_steps:
            in_memory_dicts = [config.to_dict() for config in self._in_memory_steps]
            result[TRANSFORM_IN_MEMORY_KEY] = in_memory_dicts

        if self._permanent_steps:
            permanent_dicts = [config.to_dict() for config in self._permanent_steps]
            result[TRANSFORM_PERMANENT_KEY] = permanent_dicts

        return self.transform_id, result

    @staticmethod
    def from_dict(transform_id: str,
                  transform_dict: dict):
        """
        Builds DfTransformConfig instance out of serialized dictionary.

        Parameters
        ----------
        transform_id: optional str
            Unique identifier of a transform.
        transform_dict:
            Dictionary representation of DfTransformConfig

        Returns
        -------
        DfTransformConfig instance.
        """
        in_memory_transforms = DfTransformConfig._transform_step_configs_from(transform_dict=transform_dict,
                                                                              key=TRANSFORM_IN_MEMORY_KEY)
        permanent_transforms = DfTransformConfig._transform_step_configs_from(transform_dict=transform_dict,
                                                                              key=TRANSFORM_PERMANENT_KEY)

        result = DfTransformConfig(transform_id=transform_id,
                                   in_memory_steps=in_memory_transforms,
                                   permanent_steps=permanent_transforms)
        return result

    @staticmethod
    def _transform_step_configs_from(transform_dict: dict,
                                     key: str) -> Optional[List[DfTransformStepConfig]]:
        """
        When deserializing we need to transform steps into objects as well.

        Parameters
        ----------
        transform_dict:
            Dictionary representation of DfTransformConfig
        key: str
            Key for transform steps.

        Returns
        -------
        Optional list of DfTransformStepConfig objects.
        """
        transform_dicts = transform_dict.get(key)
        transforms = None
        if transform_dicts:
            transforms = [
                DfTransformStepConfig.from_dict(transform_dict=transform_dict)
                for transform_dict in transform_dicts
            ]

        return transforms



import pandas as pd
from abc import ABC, abstractmethod
from typing import Optional


class DfCache(ABC):
    """
    Encapsulates logic how to save a dataframe to disk and how to load it afterwards.

    Init parameters
    ----------
    save_args: list
        Custom args for save method you may want to pass to your implementation.
    save_kwargs: dict
        Custom kwargs for save method you may want to pass to your implementation.
    load_args: list
        Custom args for load method you may want to pass to your implementation.
    load_kwargs: dict
        Custom kwargs for load method you may want to pass to your implementation.
    """
    def __init__(self,
                 save_args: Optional[list] = None,
                 save_kwargs: Optional[dict] = None,
                 load_args: Optional[list] = None,
                 load_kwargs: Optional[dict] = None):
        self._save_args = save_args or []
        self._save_kwargs = save_kwargs or {}
        self._load_args = load_args or []
        self._load_kwargs = load_kwargs or {}

    def save(self, df: pd.DataFrame, path: str):
        self._save(df=df, path=path, *self._save_args, **self._save_kwargs)

    def load(self, path: str) -> pd.DataFrame:
        return self._load(path=path, *self._load_args, **self._load_kwargs)

    @abstractmethod
    def _save(self, df: pd.DataFrame, path: str, *args, **kwargs):
        pass

    @abstractmethod
    def _load(self, path: str, *args, **kwargs) -> pd.DataFrame:
        pass
